# $Id: quote.rb,v 1.6 2004/07/09 09:40:42 ianmacd Exp $

require 'uri'
require 'net/http'
require 'finance'
require 'finance/currency'
require 'finance/quote/yahoo'


module Finance


  class Quote

    attr_accessor :currency, :cache

    include( *Finance::Quote::plugins )   # parentheses avoid warning

    # A wrapper class around Hash, for the purpose of returning stock data.
    #
    class QuoteData < Hash
      
      # This allows the hash to be indexed on a either a String or a Symbol
      # key.
      #
      def []( key )
	super key.to_s
      end

    end

    # Used for scaling a field. This is useful, because it handles
    # things like ranges, e.g. 105.4 - 108.3, not just set amounts.
    #
    # The method takes a String or number to scale, and the factor by which to
    # scale it. For example:
    # 
    #  scale_field( 1023, 0.01 )
    #
    # This would return <b>10.23</b>.
    #
    def Quote.scale_field( field, scale_factor )
      field.split( /([^0-9.])/ ).map do |chunk|
	chunk =~ /\d/ ? chunk.to_f * scale_factor : chunk
      end.join( '' )
    end


    # Returns a new Finance::Quote object, optionally setting the currency
    # for future method calls. If _currency_ is *nil* (the default), stock
    # prices will be returned in the native currency of their exchange.
    #
    # When _cache_ is *true* and _currency_ is not *nil*, exchange rates will
    # be cached. This represents time savings of 90%.
    #
    def initialize( currency=nil, cache=true )
      @currency = currency
      @cache = cache
    end


    # This is used to convert prices from the returned currency to the
    # user-requested currency.
    #
    def convert( data, convert_fields )

      # N.B. Caching improves look-up performance by 90%.
      @cache ? @conversion ||= {} : @conversion = {}

      # Skip all of this unless conversion is actually wanted.
      return unless @currency

      # Now process each stock in turn.
      data.keys.each do |stock|

	# Skip stocks that don't have a currency.
	stock_currency = data[stock][:currency]
	next unless stock_currency

	# Skip if it's already in the same currency.
	next if stock_currency == @currency

	# Create a Hash for the converted values.
	@conversion[stock_currency] ||= {}

	# Look up the currency conversion if we haven't already done so.
	unless @conversion[stock_currency].key?( @currency )
	  @conversion[stock_currency][@currency] =
	    Finance::Currency::convert( @currency, stock_currency )
	end

	# Make sure we have a reasonable currency conversion.
	# If we don't, mark the stock as bad.
	unless @conversion[stock_currency][@currency]
	  data[stock][:success] = false
	  data[stock][:error_msg] = 'Currency conversion failed.'
	  next
	end

	# We now have clean data, so we convert it. Ideally we'd like to just
	# *= entire fields, but unfortunately some things (like ranges,
	# capitalisation, etc.) don't take well to that. Instead, we pull out
	# any numbers we see, convert them, and stick them back in.
	convert_fields.each do |field|
	  next unless data[stock][field]

	  data[stock][field] =
	    Quote.scale_field( data[stock][field],
			       @conversion[stock_currency][@currency] )

	  # Set the new currency.
	  data[stock][:currency] = @currency
	end
      end
      
      data
    end

    private :convert

    # The hub of Ruby/Finance, this method is passed a market or _exchange_
    # from which to retrieve pricing information on _stocks_, which is a list
    # of one or more ticker symbols.
    #
    def fetch( exchange, *stocks )
      exchange = exchange.downcase.to_sym
      stocks = stocks.flatten.map { |stock| stock.upcase }

      data = QuoteData[
        send( @@delegates[exchange], stocks )
      ]

      convert( data, Finance::Currency::DEFAULT_CURRENCY_FIELDS )
      
      data
    end

  end
end
